__version__ = "0.1.0"

from pyreball.html import (set_title, print_h1, print_h2, print_h3, print_h4, print_h5, print_h6,
                           print_div, print_code, print_html, print_table, plot_graph, plot_multi_graph,
                           create_reference)
from pyreball.text import bold, em, code, ul, ol, link
